// ThreadWaitHandle.cs
//
// This file contains a class called ThreadWaitHandle that is a derivative
// of System.Threading.WaitHandle and that can be used to perform
// WaitHandle.Wait* operations on thread objects.  The purpose of the code
// is to allow a programmer to create one or more threads and then, at some
// later point, block until that thread exits.
//
// Some notes about this implementation:
//
// o This code assumes that each CLR Thread object is implemented on top
//   of a native Win32 thread object.
//
// o This code uses reflection and unsafe pointer arithmetic to extract
//   the native thread HANDLE using knowledge gained by observing where
//   the native Win32 thread HANDLE is stored in memory referred to by
//   the CLR Thread object using the DONT_USE_InternalThread field.
//
// As such, this code is highly coupled to the build of the CLR that
// the code was developed on.  Any change to how CLR Thread object's
// map onto Win32 thread objects, or any change to where the native Win32
// HANDLE is stored in memory, will invalidate this code.
//
// Also note that if you're creating threads that have very short lifetimes,
// it's very possible that the thread you create could start, do it's work,
// and exit before you have a chance to return from calling Thread.Start
// and subsequently construct an instance of the ThreadWaitHandle class.
// If this happens, debug builds of this code will incur an assertion failure.
// If this happens to you, you'll need to use a thread synchronization
// technique (like an event) to coordinate the startup of the new thread
// so that it doesn't proceed to do it's work until you've had a chance to
// instantiate the ThreadWaitHandle class and extract the native thread handle.
//
// For example, assuming the existence of an auto-reset event named "startEvent"
// that has already been initialized to the reset state, the following code
// would be used in the 'creating' thread:
//
//  // Start the new thread:
//  //
//  Thread t = new Thread(new ThreadStart(YourThreadProc));
//  t.Start();
// 
//  // Initialize the wait handle for this thread.
//  //
//  ThreadWaitHandle twh = new ThreadWaitHandle(t);
//
//  // Tell the worker thread that it's okay to proceed.
//  //
//  startEvent.Set();
//
// While the worker thread would look like this:
//
//  void YourThreadProc()
//  {
//      // Wait until we've been command to actually
//      // start doing our thing...
//      //
//      startEvent.WaitOne();
//
//      // ...normal thread code here...
//  }
//
// Thanks to Howard Hoffman for reporting the problem to my attention, which
// led to identifying the race condition between initializing the ThreadWaitHandle
// for very short-lived threads.
//
// Mike Woodring
// http://staff.develop.com/woodring
//
using System;
using System.Threading;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Diagnostics;

public class ThreadWaitHandle : WaitHandle
{
    // _handleOffset
    //
    // This value represents the index into the native
    // thread information structure where the native thread
    // handle is stored (in terms of DWORD offsets).  For example,
    // a value of 43 means the 43rd 32-bit value holds the
    // native thread handle.
    //
    // Due to the highly coupled nature of this information to
    // the runtime, if the current runtime doesn't exactly
    // match one of the known, tested versions and exception
    // will be raised if this class is subsequently used.
    //
    // Since this offset will be constant at runtime for a given
    // version of the CLR, it is computed only once when the
    // ThreadWaitHandle class is first used in an AppDomain.
    //
    static readonly int s_handleOffset = -1;

    // s_nativeThreadFieldInfo
    //
    // This is the reflection information for the "DONT_USE_InternalThread"
    // field of the CLR managed Thread class.  It has been observed that
    // this field holds a pointer to the native structure representing
    // the thread, of which the native thread handle is one element (as
    // described above).
    //
    // As with s_handleOffset, this information will hold constant
    // in memory for a given run of the program, so it's just looked up
    // once when this class is first used in an AppDomain.
    //
    // If this field can no longer be found, an exception will be raised,
    // but not unless this class is actually used.
    //
    static readonly FieldInfo s_nativeThreadFieldInfo = null;
    
    struct VersionToHandleOffsetMapping
    {
        internal readonly Version runtimeVersion;
        internal readonly int     handleOffset;

        internal VersionToHandleOffsetMapping( Version v, int o )
        {
            runtimeVersion = v;
            handleOffset = o;
        }
    }

    static VersionToHandleOffsetMapping[] s_versionToHandleOffsetMap =
    {
        // The following has been observed for SP1 of the .NET Framework:
        //
        // A Thread instance has a private IntPtr field (DONT_USE_InternalThread)
        // that is a pointer to a structure describing the thread.  The HANDLE
        // to the native Win32 thread object is stored at the 32bit field located
        // at byte offset 0xAC from the top of that structure (the 43rd
        // dword).
        //
        // For build 1.1.4322.510 of the runtime (the VS.NET 2003 "final beta"),
        // the handle is stored at the 45th DWORD.
        //
        new VersionToHandleOffsetMapping(new Version(1, 0, 3705, 288), 43),     // 1.0 + SP1
        new VersionToHandleOffsetMapping(new Version(1, 1, 4322, 510), 45)      // 1.1 "final beta"
    };

    static ThreadWaitHandle()
    {
        // Cross-reference the current version of the runtime against
        // the map of known versions-to-handle-offsets.
        //
        Version currentRuntime = Environment.Version;

        foreach( VersionToHandleOffsetMapping mapping in s_versionToHandleOffsetMap )
        {
            if( currentRuntime == mapping.runtimeVersion )
            {
                s_handleOffset = mapping.handleOffset;
                break;
            }
        }

        // If we don't find a matching runtime, s_handleOffset
        // will still be -1, which we'll check for below.

        // Lookup the FieldInfo for the internal native thread
        // information field.  If this field no longer exists,
        // s_nativeThreadFieldInfo will be null, which we'll check
        // for below.
        //
        s_nativeThreadFieldInfo =
            typeof(Thread).GetField( "DONT_USE_InternalThread",
                                     BindingFlags.Instance | BindingFlags.NonPublic );
    }

    public ThreadWaitHandle()
        : this(Thread.CurrentThread)
    {
    }

    public ThreadWaitHandle( Thread targetThread )
    {
        if( s_handleOffset == -1 )
        {
            string errorMsg = string.Format( "The ThreadWaitHandle class has not been tested on the current " +
                                             "version of the runtime ({0}).  The location of the native thread handle " +
                                             "needs to be re-discovered and the version-to-handle-offset-map updated accordingly.",
                                             Environment.Version );

            throw new ApplicationException(errorMsg);
        }

        if( s_nativeThreadFieldInfo == null )
        {
            string errorMsg = string.Format( "The required DONT_USE_InternalThread field of the managed " +
                                            "Thread class no longer exists, so the ThreadWaitHandle class " +
                                            "is not functional on this version of the runtime ({0}).",
                                             Environment.Version );

            throw new ApplicationException(errorMsg);
        }

        // Grab the native thread handle, duplicating it to update
        // the reference count on the thread, and stash it in the
        // base class WaitHandle's Handle property.  When this
        // WaitHandle instance is finalized, disposed, or closed,
        // the base class will call the Win32 CloseHandle function
        // to reduce the reference count accordingly.
        //
        Handle = GetNativeHandle(targetThread, true);
    }

    // Return a non-addref'd handle to the current thread.
    //
    public static IntPtr GetNativeHandle()
    {
        return GetNativeHandle(Thread.CurrentThread);
    }

    // Return a possibly addref'd handle to the current thread.
    //
    public static IntPtr GetNativeHandle( bool addRef )
    {
        return GetNativeHandle(Thread.CurrentThread, addRef);
    }

    // Return a non-addref'd handle to the specified thread.
    //
    public static IntPtr GetNativeHandle( Thread targetThread )
    {
        return GetNativeHandle(targetThread, false);
    }

    // This following 2 helper functions can be used to increment
    // and decrement the OS reference count for the specified native
    // handle.
    //
    public static IntPtr AddRef( IntPtr nativeHandle )
    {
        if( nativeHandle == InvalidHandle )
        {
            throw new ArgumentException("Cannot add a reference to an invalid handle.", "nativeHandle");
        }

        return new IntPtr(DuplicateHandle(nativeHandle.ToInt32()));
    }

    public static void Release( IntPtr nativeHandle )
    {
        if( nativeHandle == InvalidHandle )
        {
            throw new ArgumentException("Cannot release a reference to an invalid handle.", "nativeHandle");
        }

        CloseHandle(nativeHandle.ToInt32());
    }

    // Return a posssibly addref'd handle to the specified thread.
    //
    public unsafe static IntPtr GetNativeHandle( Thread targetThread, bool addRef )
    {
        try
        {
            // Fetch the value of that field and grab the dword at the previously
            // determined offset (s_handleOffset) where the native thread handle
            // is stored.
            //
            IntPtr ThreadInfoPtr = (IntPtr)s_nativeThreadFieldInfo.GetValue(targetThread);
            int hThread = ((int *)ThreadInfoPtr.ToPointer())[s_handleOffset];

            Debug.Assert( (hThread != InvalidHandle.ToInt32()) && (hThread != 0),
                          "(hThread != InvalidHandle) && (hThread != 0)",
                          "The specified target thread does not have a valid handle.  This typically means it either  " +
                          "hasn't been started yet, that it has already exited, or that this code is running on an " +
                          "untested version of the runtime." );

            if( addRef )
            {
                // Return an add-ref'd handle to this thread.  If the return
                // value is being stored in a WaitHandle-derived object as the
                // native handle, then the WaitHandle base class's implementation
                // of Close/Dispose will take care of calling the Win32
                // CloseHandle function to release this reference to the
                // underlying thread object.
                //
                hThread = DuplicateHandle(hThread);
            }

            return new IntPtr(hThread);
        }
        catch
        {
            return(InvalidHandle);
        }
    }

    [ DllImport("kernel32.dll") ]
    public static extern bool CloseHandle( int h );

    [ DllImport("kernel32.dll") ]
    static extern int GetCurrentProcess();

    [ DllImport("kernel32.dll") ]
    static extern bool DuplicateHandle( int hSourceProcess, int originalHandle,
                                        int hTargetProcess, out int newHandle,
                                        uint accessFlags, bool inheritHandle,
                                        uint options );

    const int DUPLICATE_SAME_ACCESS = 0x00000002; // From WINNT.H.

    static int DuplicateHandle( int h )
    {
        int hCurProcess = GetCurrentProcess();
        int newHandle;
        
        if( DuplicateHandle( hCurProcess, h, hCurProcess, out newHandle,
                             0, false, DUPLICATE_SAME_ACCESS) )
        {
            return(newHandle);
        }
        else
        {
            return InvalidHandle.ToInt32();
        }
    }
}
